#! /usr/bin/perl

##############################################################################
#
#  msacs-eval.pl : evaluation program for the SemEval-2007 task#9
#                  based on the evaluation program for CoNLL-2005 Shared Task
#
#  Authors : Luis Villarejo
#  Contact : luisv at lsi.upc.edu
#
#  Created : March 2007 on top of srlconll 2005 srl-eval script (Carreras)
#  Modified: -
#
###############################################################################


use strict;

use MSACS::sentence;
use MSACS::phraseset;
use MSACS::prop;


############################################################
#  A r g u m e n t s   a n d   H e l p

use Getopt::Long; 
my %options;
GetOptions(\%options, 
           "latex",     # latex output
	   "C",         # confusion matrix
	   "noW",
	   "NSD",
	   "NER",
	   "SC",
	   "PROPS"
           ); 


my $script = "msacs-eval.pl"; 
my $help = << "end_of_help;";
Usage: perl msacs-eval.pl <annotation to evaluate> [<options>] <basic info file> <gold props> <predicted props>

     <annotation to evaluate> can be either:
        -NER      Named entities
        -NSD      Noun Senses
        -SC       Semantic Classes
        -PROPS    Semantic Roles

Options:

     -latex      Produce a results table in LaTeX
     -C          Produce a confusion matrix of gold vs. predicted argments, wrt. their role

Example:

     perl msacs-eval.pl -NSD train.ca.BII.txt train.ca.NSD.txt train.ca.NSD.txt.prediction
     perl msacs-eval.pl -PROPS train.ca.BII.txt train.ca.SRL.txt train.ca.SRL.txt.prediction
     perl msacs-eval.pl -SC train.ca.BII.txt train.ca.SRL.txt train.ca.SRL.txt.prediction

end_of_help;

############################################################
# P R E   P R O C E S S I N G

# open files and create eval files

if (@ARGV < 3) {
    print $help;
    exit;
}

my $basicfile = shift @ARGV;
my $preGoldfile = shift @ARGV;
my $prePredfile = shift @ARGV;

my $goldfile;
my $predfile;

my $column;
if ($options{PROPS}){  $column = "SRL"; } # TV + PROPS as they are
elsif($options{SC}) {  $column = "SC";  } # TV + (SC*)
elsif($options{NSD}){  $column = "NSD"; } # TN + (NE*)
elsif($options{NER}){  $column = "NER"; } # The target column is generated from gold file
else{  die $help;  }

if ($preGoldfile =~ /^(.+)\.gz$/) {
    $goldfile = $1.".evalGold.$column";
    open PREGOLD, "gunzip -c $preGoldfile |" or die "Could not open gzipped file of gold props ($preGoldfile)! $!\n";
}
else {
    $goldfile = $preGoldfile.".evalGold.$column";
    open PREGOLD, $preGoldfile or die "Could not open file of gold props ($preGoldfile)! $!\n"; 
}
if ($predfile =~ /\.gz/) {
    $predfile = $1.".evalPred.$column";
    open PREPRED, "gunzip -c $prePredfile |" or die "Could not open gzipped file of predicted props ($prePredfile)! $!\n";
}
else {
    $predfile = $prePredfile.".evalPred.$column";
    open PREPRED, $prePredfile or die "Could not open file of prePredicted props ($prePredfile)! $!\n"; 
}

buildFiles($column, $basicfile, $goldfile, $predfile, PREGOLD => \*PREGOLD, PREPRED => \*PREPRED);



############################################################
#  M A I N   P R O G R A M 


my $ns = 0;        # number of sentence
my $ntargets = 0;  # number of target verbs
my %E;             # evaluation results
my %C;             # confusion matrix

my %excluded = ( V => 1); 

# open files

open GOLD, $goldfile or die "$script: could not open file of gold props ($goldfile)! $!\n";
open PRED, $predfile or die "$script: could not open file of predicted props ($predfile)! $!\n";

##
# read and evaluate propositions, sentence by sentence

my $s = MSACS::sentence->read_props($ns, GOLD => \*GOLD, PRED => \*PRED, HASH => \%options);

while ($s) {

    my $prop; 

    my (@G, @P, $i); 
    
    map { $G[$_->position] = $_ } $s->gold_props;
    map { $P[$_->position] = $_ } $s->pred_props;

    for($i=0; $i<@G; $i++) {
	my $gprop = $G[$i];
	my $pprop = $P[$i];

	if ($pprop and !$gprop) {	
	    !$options{noW} and print STDERR "WARNING : sentence $ns : verb ", $pprop->verb, 
	    " at position ", $pprop->position, " : found predicted prop without its gold reference! Skipping prop!\n";
	}
	elsif ($gprop) {
	    if (!$pprop) {
		!$options{noW} and print STDERR "WARNING : sentence $ns : verb ", $gprop->verb, 
		" at position ", $gprop->position, " : missing predicted prop! Counting all arguments as missed!\n";
		$pprop = MSACS::prop->new($gprop->verb, $gprop->position); 
	    }
	    elsif ($gprop->verb ne $pprop->verb) {
		!$options{noW} and  print STDERR "WARNING : sentence $ns : props do not match : expecting ",
		$gprop->verb, " at position ", $gprop->position, 
		", found ", $pprop->verb, " at position ", $pprop->position, "! Counting all gold arguments as missed!\n";
		$pprop = MSACS::prop->new($gprop->verb, $gprop->position);
	    }
	
	    $ntargets++;
	    my %e = evaluate_proposition($gprop, $pprop);


	    # Update global evaluation results

	    $E{ok} += $e{ok};
	    $E{op} += $e{op};
	    $E{ms} += $e{ms};
	    $E{ptv} += $e{ptv};

	    my $t; 
	    foreach $t ( keys %{$e{T}} ) {
		$E{T}{$t}{ok} += $e{T}{$t}{ok};
		$E{T}{$t}{op} += $e{T}{$t}{op};
		$E{T}{$t}{ms} += $e{T}{$t}{ms};
	    }
	    foreach $t ( keys %{$e{E}} ) {
		$E{E}{$t}{ok} += $e{E}{$t}{ok};
		$E{E}{$t}{op} += $e{E}{$t}{op};
		$E{E}{$t}{ms} += $e{E}{$t}{ms};
	    }

	    if ($options{C}) {
		update_confusion_matrix(\%C, $gprop, $pprop); 
	    }
	}
    }

    $ns++;
    $s = MSACS::sentence->read_props($ns, GOLD => \*GOLD, PRED => \*PRED); 

}

system("rm $goldfile");
system("rm $predfile");

# Print Evaluation results
my $t; 

if ($options{latex}) {
#    print '\begin{table}[t]', "\n"; 
#    print '\centering', "\n"; 
    print '\begin{tabular}{|l|r|r|r|}\cline{2-4}',  "\n";
    print '\multicolumn{1}{c|}{datalabel}', "\n";
    print '           & Precision & Recall & F$_{\beta=1}$', '\\\\', "\n", '\hline', "\n";  #'

    printf("%-18s & %6.2f\\%% & %6.2f\\%% & %6.2f\\\\\n", "Overall", precrecf1($E{ok}, $E{op}, $E{ms}));
    print '\hline', "\n"; 

    foreach $t ( sort keys %{$E{T}} ) {
	printf("%-18s & %6.2f\\%% & %6.2f\\%% & %6.2f\\\\\n", "\\texttt{".$t."}", precrecf1($E{T}{$t}{ok}, $E{T}{$t}{op}, $E{T}{$t}{ms}));
    }
    print '\hline', "\n";

    if (%excluded) {
	print '\hline', "\n"; 
	foreach $t ( sort keys %{$E{E}} ) {
	    printf("%-18s & %6.2f\\%% & %6.2f\\%% & %6.2f\\\\\n", "\\texttt{".$t."}", precrecf1($E{E}{$t}{ok}, $E{E}{$t}{op}, $E{E}{$t}{ms})); 
	}
	print '\hline', "\n"; 
    }
    
    print '\end{tabular}', "\n"; 
#    print '\end{table}', "\n"; 
}
else {
    printf("Number of Sentences    :      %6d\n", $ns); 
    printf("Number of Propositions :      %6d\n", $ntargets); 
    printf("Percentage of perfect props : %6.2f\n",($ntargets>0 ? 100*$E{ptv}/$ntargets : 0)); 
    print "\n"; 
    
    printf("%10s   %6s  %6s  %6s   %6s  %6s  %6s\n", "", "corr.", "excess", "missed", "prec.", "rec.", "F1"); 
    print "------------------------------------------------------------\n"; 
    printf("%10s   %6d  %6d  %6d   %6.2f  %6.2f  %6.2f\n", 
	   "Overall", $E{ok}, $E{op}, $E{ms}, precrecf1($E{ok}, $E{op}, $E{ms})); 
#    print "------------------------------------------------------------\n"; 
    print "----------\n"; 
    
#    printf("%10s   %6d  %6d  %6d   %6.2f  %6.2f  %6.2f\n", 
#	   "all - {V}", $O2{ok}, $O2{op}, $O2{ms}, precrecf1($O2{ok}, $O2{op}, $O2{ms})); 
#    print "------------------------------------------------------------\n"; 
    
    foreach $t ( sort keys %{$E{T}} ) {
	printf("%10s   %6d  %6d  %6d   %6.2f  %6.2f  %6.2f\n", 
	       $t, $E{T}{$t}{ok}, $E{T}{$t}{op}, $E{T}{$t}{ms}, precrecf1($E{T}{$t}{ok}, $E{T}{$t}{op}, $E{T}{$t}{ms})); 
    }
    print "------------------------------------------------------------\n"; 

   foreach $t ( sort keys %{$E{E}} ) {
	printf("%10s   %6d  %6d  %6d   %6.2f  %6.2f  %6.2f\n", 
	       $t, $E{E}{$t}{ok}, $E{E}{$t}{op}, $E{E}{$t}{ms}, precrecf1($E{E}{$t}{ok}, $E{E}{$t}{op}, $E{E}{$t}{ms})); 
    }
    print "------------------------------------------------------------\n"; 
}


# print confusion matrix
if ($options{C}) {

    my $k; 

    # Evaluation of Unlabelled arguments
    my ($uok, $uop, $ums, $uacc) = (0,0,0,0); 
    foreach $k ( grep { $_ ne "-NONE-" && $_ ne "V" } keys %C ) {	
	map { $uok += $C{$k}{$_} } grep { $_ ne "-NONE-" && $_ ne "V" } keys %{$C{$k}};
	$uacc += $C{$k}{$k};
	$ums += $C{$k}{"-NONE-"};
    }
    map { $uop += $C{"-NONE-"}{$_} } grep { $_ ne "-NONE-" && $_ ne "V" } keys %{$C{"-NONE-"}};

    print "--------------------------------------------------------------------\n"; 
    printf("%10s   %6s  %6s  %6s   %6s  %6s  %6s  %6s\n", "", "corr.", "excess", "missed", "prec.", "rec.", "F1", "lAcc"); 
    printf("%10s   %6d  %6d  %6d   %6.2f  %6.2f  %6.2f  %6.2f\n", 
	   "Unlabeled", $uok, $uop, $ums, precrecf1($uok, $uop, $ums), 100*$uacc/$uok); 
    print "--------------------------------------------------------------------\n"; 



    print "\n---- Confusion Matrix: (one row for each correct role, with the distribution of predictions)\n"; 

    my %AllKeys; 
    map { $AllKeys{$_} = 1 } map { $_, keys %{$C{$_}} } keys %C;
    my @AllKeys = sort keys %AllKeys; 


    
    my $i = -1; 
    print "             ";
    map { printf("%4d ", $i); $i++} @AllKeys;
    print "\n";
    $i = -1;
    foreach $k ( @AllKeys ) {
	printf("%2d: %-8s ", $i++, $k); 
	map { printf("%4d ", $C{$k}{$_}) } @AllKeys;
	print "\n"; 
    }


    my ($t1,$t2); 
    foreach $t1 ( sort keys %C ) {
	foreach $t2 ( sort keys %{$C{$t1}} ) {
#	    printf("    %-6s vs %-6s :  %-5d\n", $t1, $t2, $C{$t1}{$t2});
	}
    }    
}

# end of main program
#####################

############################################################
#  S U B R O U T I N E S


# evaluates a predicted proposition wrt the gold correct proposition
# returns a hash with the following keys
#   ok  :  number of correctly predicted args
#   ms  :  number of missed args
#   op  :  number of over-predicted args
#   T   :  a hash indexed by argument types, where
#           each value is in turn a hash of {ok,ms,op} numbers
#   E   :  a hash indexed by excluded argument types, where
#           each value is in turn a hash of {ok,ms,op} numbers
sub evaluate_proposition {
    my ($gprop, $pprop) = @_;

    my $o = $gprop->discriminate_args($pprop); 
    
    my %e; 

    my $a; 
    foreach $a (@{$o->{ok}}) {
	if (!$excluded{$a->type}) {
	    $e{ok}++;
	    $e{T}{$a->type}{ok}++;
	}
	else {
	    $e{E}{$a->type}{ok}++;
	}
    }
    foreach $a (@{$o->{op}}) {
	if (!$excluded{$a->type}) {
	    $e{op}++;
	    $e{T}{$a->type}{op}++;
	}
	else {
	    $e{E}{$a->type}{op}++;
	}
    }
    foreach $a (@{$o->{ms}}) {
	if (!$excluded{$a->type}) {
	    $e{ms}++;
	    $e{T}{$a->type}{ms}++;
	}
	else {
	    $e{E}{$a->type}{ms}++;
	}
    }

    $e{ptv} = (!$e{op} and !$e{ms}) ? 1 : 0; 
    
    return %e; 
}


# computes precision, recall and F1 measures
sub precrecf1 {
    my ($ok, $op, $ms) = @_;

    my $p = ($ok + $op > 0) ? 100*$ok/($ok+$op) : 0; 
    my $r = ($ok + $ms > 0) ? 100*$ok/($ok+$ms) : 0; 

    my $f1 = ($p+$r>0) ? (2*$p*$r)/($p+$r) : 0; 

    return ($p,$r,$f1); 
}




sub update_confusion_matrix {
    my ($C, $gprop, $pprop) = @_;

    my $o = $gprop->discriminate_args($pprop, 0); 

    my $a; 
    foreach $a ( @{$o->{ok}} ) {
	my $g = shift @{$o->{eq}};
	$C->{$g->type}{$a->type}++;
    }
    foreach $a ( @{$o->{ms}} ) {
	$C->{$a->type}{"-NONE-"}++;
    }
    foreach $a ( @{$o->{op}} ) {
	$C->{"-NONE-"}{$a->type}++;
    }
}

# Builds the proper files for the scorer to run.
sub buildFiles{
    my ($column, $basicfile, $goldfile, $predfile, %FILES) = @_;
    my $command;
    my $printFormat;

    open(PRED, ">$predfile") or die "Could not write to file $predfile! $!\n";
    open(GOLD, ">$goldfile") or die "Could not write to file $goldfile! $!\n";

    if ( ($column eq "NSD") || ($column eq "NER") ){  $command = "\$1\" \"\$2";  }
    elsif ( ($column eq "SRL") || ($column eq "SC") ){  $command = "\$1\" \"\$3";  }

    open TN, "awk \'{print $command}\' $basicfile |" or die "$script: can't extract targets column from $basicfile $!\n";

    if ( exists($FILES{PREGOLD}) && exists($FILES{PREPRED}) ) {
      my $pregoldFile = $FILES{PREGOLD};
      my $prepredFile = $FILES{PREPRED};

      while ( my $lineTN = <TN>)  {
	my $linePG = <$pregoldFile>;
	my $linePP = <$prepredFile>;
	chomp($lineTN);
	chomp($linePG);
	chomp($linePP);

	if ($column eq "NSD"){

	  $linePG =~ s/^.*([0-9]{8}.).*$/\($1\*\)/g;
	  $linePG =~ s/^.*-.*$/*/g;
	  $linePG = sprintf('%11s', $linePG);

	  $linePP =~ s/^.*([0-9]{8}.).*$/\($1\*\)/g;
	  $linePP =~ s/^.*-.*$/*/g;
	  $linePP = sprintf('%11s', $linePP);
	}
	elsif ($column eq "SRL"){

	  my @partsPG = split(/[ +\t]/, $linePG);
	  my $linePGtmp;
	  for(my $i = 1; $i < scalar(@partsPG); $i++){ # Starting at one we avoid the SC column
	    $linePGtmp .= sprintf('%11s  ', $partsPG[$i]);
	  }
	  $linePG = $linePGtmp;

	  my @partsPP = split(/[ +\t]/, $linePP);
	  my $linePPtmp;
	  for(my $i = 1; $i < scalar(@partsPP); $i++){ # Starting at one we avoid the SC column
    	      $linePPtmp .= sprintf('%11s  ', $partsPP[$i]);
	  }
	  $linePP = $linePPtmp;

	}
	elsif ($column eq "SC"){
	  my @partsPG = split(/[ +\t]/, $linePG);
	  $partsPG[0] =~ s/^([a-z|A-Z][0-9])$/\($1\*\)/g;
	  $partsPG[0] = lc($partsPG[0]);
	  $partsPG[0] =~ s/^.*-.*$/*/g;
	  $linePG = sprintf('%2s  ', $partsPG[0]);

	  my @partsPP = split(/[ +\t]/, $linePP);
	  $partsPP[0] =~ s/^([a-z|A-Z][0-9])$/\($1\*\)/g;
	  $partsPP[0] = lc($partsPP[0]);
	  $partsPP[0] =~ s/^.*-.*$/*/g;
	  $linePP = sprintf('%2s  ', $partsPP[0]);
	}
	
	my $word = " ";
	if ($lineTN !~ /^\s$/){
	  my @parts = split(/[ +\t]/, $lineTN);
	  if ( ($column eq "NER") || ($column eq "NSD") ){ #Target guided by gold
	    if ($linePG =~ /\(/){  $word = sprintf('%-18s', $parts[0]);  }
	    else{ $word = sprintf('%-18s', "-"); }
	  }
	  else{
	    if ($parts[1] eq "*"){  $word = sprintf('%-18s', $parts[0]);  }
	    else{  $word = sprintf('%-18s', "-");  }	
	  }
	}
	
	#print STDERR "w: $word  \nPG: $linePG\nPP: $linePP\n";getc();
	print PRED "$word  $linePP\n"; # printf PRED ($printFormat, $word, $linePP);
	print GOLD "$word  $linePG\n"; # printf GOLD ($printFormat, $word, $linePG);
      }

    }
    else{  die "Can't open either PREGOLD or PREPRED file.\n";  }

    close TN;
    close PRED;
    close GOLD;

  }

# end of script
###############





















